package org.xbib.interlibrary.client.bvb.z;

import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.common.MultiMap;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.action.avail.AvailRequest;
import org.xbib.interlibrary.action.avail.AvailResponse;
import org.xbib.interlibrary.action.avail.AvailResponseBuilder;
import org.xbib.interlibrary.action.avail.AvailServiceResponseBuilder;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.library.LibraryService;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.library.DefaultLibraryService;
import org.xbib.interlibrary.api.action.Service;
import org.xbib.interlibrary.api.action.ServiceArguments;
import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.api.library.Mode;
import org.xbib.interlibrary.z.AbstractZAvailService;
import org.xbib.interlibrary.z.DisplayHelper;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecord;
import org.xbib.z3950.client.api.Client;
import static org.xbib.interlibrary.api.library.Material.ONLINE_RESOURCE;

public class ZAvailService extends AbstractZAvailService {

    private static final Logger logger = Logger.getLogger(ZAvailService.class.getName());

    public ZAvailService(ServiceArguments arguments) {
        super(arguments);
    }

    @Override
    protected String supports(String source) {
        return source.toLowerCase(Locale.ROOT).contains("bvb") ? "BVB" : null;
    }

    @Override
    protected void fetchAvailServices(AvailRequest availRequest,
                                      AvailServiceResponseBuilder availResponse,
                                      MultiMap<String, LibraryService> multiMap) {
        try {
            lock.lock();
            String query = getResolvePQF(availRequest);
            if (query == null) {
                return;
            }
            logger.log(Level.FINE, () -> "executing avail request = " + availRequest + " pqf = '" + query + "'");
            zClientHelper.open();
            final Client client = zClientHelper.getClient();
            int count = client.searchPQF(query, 0, 1, null,
                    (status, total, returned, elapsedMillis) ->
                            logger.log(Level.FINE, "status = " + status + " total = " + total + " returned = " + returned + " millis = " + elapsedMillis),
                    record -> {
                        try (InputStream inputStream = record.asStream()) {
                            String format = zClientHelper.getFormat();
                            String type = zClientHelper.getType();
                            Charset charset = Charset.forName(zClientHelper.getEncoding());
                            boolean found = false;
                            Marc.Builder builder = Marc.builder()
                                    .setFormat(format)
                                    .setType(type)
                                    .setInputStream(inputStream)
                                    .setCharset(charset);
                            for (MarcRecord marcRecord : builder.iterable()) {
                                avail(availRequest, availResponse, multiMap, marcRecord);
                            }
                        } catch (Exception e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                            availResponse.setStatus(getName(), Response.Status.ERROR);
                            availResponse.setThrowable(getName(), e.getMessage(), e);
                            throw new IllegalStateException(e);
                        }
                    },
                    () -> availResponse.setStatus(getName(), Response.Status.TIMEOUT));
            zClientHelper.close();
            logger.log(Level.FINE, () -> "result count = " + count);
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            availResponse.setStatus(getName(), Response.Status.ERROR);
            availResponse.setThrowable(getName(), e.getMessage(), e);
            throw new IllegalStateException(e);
        } finally {
            lock.unlock();
        }
    }

    @Override
    protected void avail(AvailRequest availRequest,
                         AvailServiceResponseBuilder availResponse,
                         MultiMap<String, LibraryService> multiMap,
                         MarcRecord marcRecord) {
        boolean found = false;
        if (availRequest.isBibliographicDescriptionEnabled()) {
            availResponse.setBibliographicDescription(marcRecord);
            TinyMap.Builder<String, Object> builder = TinyMap.builder();
            DisplayHelper.display(builder, marcHelper, marcRecord, getName(), getName().toUpperCase(Locale.ROOT), null,
                    marcHelper.getRecordIdSupplier(marcRecord));
            availResponse.setDisplay(builder.build());
        }
        for (LibraryService service : extractBvbServices(availRequest, marcRecord)) {
            multiMap.put(service.getISIL(), service);
            found = true;
        }
        if (found) {
            availResponse.setStatus(getName(), Response.Status.OK);
        } else {
            availResponse.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
        }
    }

    @Override
    public int compareTo(Service<AvailRequest, AvailResponse, AvailResponseBuilder> o) {
        return getName().compareTo(o.getName());
    }

    private List<LibraryService> extractBvbServices(AvailRequest availRequest,
                                                    MarcRecord marcRecord) {
        List<Material> materials = marcHelper.extractMaterial(marcRecord);
        List<LibraryService> list = new ArrayList<>();
        String identifier = identifier(marcRecord);
        marcRecord.getAll("049").forEach(marcField ->
                marcField.getSubfields().forEach(subfield -> {
                    String isil = subfield.getValue();
                    if (isil != null) {
                        Library library = configuration.getLibrary(isil);
                        if (library != null) {
                            DefaultLibraryService.Builder builder = DefaultLibraryService.builder();
                            builder.setISIL(library.getISIL());
                            builder.setName(library.getName());
                            builder.setDomainPriority(domainPriority(availRequest, library));
                            builder.setLibraryPriority(libraryPriority(availRequest, library));
                            builder.setMaterials(materials);
                            builder.setServicePriority(materials.contains(ONLINE_RESOURCE) ? 1 : 2);
                            builder.setSource(library.getDomainName());
                            builder.setSourceId(identifier);
                            // always assume loan and copy
                            builder.setMode(Mode.LOAN_COPY);
                            list.add(builder.build());
                        } else {
                            logger.log(Level.WARNING, "ISIL not found in configuration: " + isil);
                        }
                    }
                })
        );
        return list;
    }
}
